import mimetypes
import typing as tp
import os
import random
import string
from email.header import Header
import uuid
from django import forms
from django.core.validators import URLValidator
from django.forms import fields
from django.core.exceptions import ValidationError
from django.http import Http404, StreamingHttpResponse
from django.shortcuts import render, get_object_or_404, redirect
from django.contrib.auth.decorators import login_required
from django.utils.safestring import mark_safe

from shares.models import Share, SHARE_URL, SHARE_FILE
import hashlib
from netguru.settings import SECRET_KEY, FILE_SAVE_LOCATION


def hash_password(password: str) -> str:
    return hashlib.sha256(password.encode('utf-8') + SECRET_KEY.encode('utf-8')).hexdigest()


def generate_correct_uuid() -> str:
    f = uuid.uuid4().hex
    # I know that hash collisions practically don't happen, but since it's so cheap for us
    # to go that extra mile
    while os.path.exists(FILE_SAVE_LOCATION, f):
        f = uuid.uuid4().hex
    return f

# Cut off O, I and S so that the user doesn't confuse them with 0, 1 and 5 respectively
CHARACTERS_TO_USE = string.ascii_uppercase.replace('O', '') \
                          .replace('I', '').replace('S', '') + string.digits


def random_password() -> str:
    """Generate a eight character reasonably safe password"""
    return ''.join(random.choice(CHARACTERS_TO_USE) for i in range(6))


class AddShareURLForm(forms.Form):
    url = fields.CharField(label='URL', required=False,
                           validators=[URLValidator()])
    file = fields.FileField(label='File', required=False)

    def clean(self):
        cleaned_data = super().clean()
        if not cleaned_data.get('url') and not cleaned_data.get('file'):
            raise ValidationError('You have to provide either an URL or a file!')
        return cleaned_data


@login_required
def add_share(request):
    data_added = None

    if request.method == 'POST':
        form = AddShareURLForm(request.POST, request.FILES)
        if form.is_valid():
            data_added = {'password': random_password()}
            pwd_hash = hash_password(data_added['password'])
            data = form.cleaned_data
            if data.get('url'):
                # Create a URL resource
                share = Share(creator=request.user,
                              resource=data['url'],
                              pwd_hash=pwd_hash,
                              share_type=SHARE_URL)
                share.save()
                data_added['url'] = mark_safe(f'https://{request.get_host()}/shares/{share.id}')
            else:
                file_name = generate_correct_uuid()
                file = request.FILES['file']
                resource_name = f'{file_name}.{file.name}'
                with open(os.path.join(FILE_SAVE_LOCATION, file_name), 'wb') as f_out:
                    for chunk in file.chunks():
                        f_out.write(chunk)
                share = Share(creator=request.user,
                              resource=resource_name,
                              pwd_hash=pwd_hash,
                              share_type=SHARE_FILE)
                share.save()
                data_added['url'] = mark_safe(f'https://{request.get_host()}/shares/{share.id}')
    else:
        form = AddShareURLForm()

    return render(request, 'share/add.html', {'form': form, 'added': data_added})


class PasswordForm(forms.Form):
    password = fields.CharField(label='Password', widget=forms.PasswordInput)


def view_share(request, share_id: str):
    share = get_object_or_404(Share, id=share_id)
    if request.method == 'POST':
        form = PasswordForm(request.POST)
        if form.is_valid():
            pwd = form.cleaned_data['password']
            if hash_password(pwd) == share.pwd_hash:
                try:
                    if share.share_type == SHARE_URL:
                        return redirect(share.resource)
                    else:
                        file_uuid, file_name = share.resource.split('.', 1)
                        file_loc = os.path.join(FILE_SAVE_LOCATION, file_uuid)
                        if not os.path.exists(file_loc):
                            raise Http404()
                        else:
                            def file_iterator(path: str) -> tp.Iterator[bytes]:
                                with open(path, 'rb') as f:
                                    p = f.read(1024)
                                    while p:
                                        yield p
                                        p = f.read(1024)

                            # Django will have the UTF-8 mumbo jumbo for us
                            cd = f'attachment; filename="{file_name}"'
                            mimetype = mimetypes.guess_type(file_name)[0] or 'application/octet-stream'
                            return StreamingHttpResponse(file_iterator(file_loc),
                                                         content_type=mimetype,
                                                         headers={
                                                             'Content-Disposition': cd
                                                         })
                finally:
                    share.views_count += 1
                    share.save()
    else:
        form = PasswordForm()

    return render(request, 'share/view.html', {'form': form, 'share_id': share_id})
